import json
import logging
import os
from pathlib import Path
from typing import Dict
from olot.oci.oci_config import OCIManifestConfig
from olot.oci.oci_image_index import OCIImageIndex
from olot.utils.types import compute_hash_of_str
from olot.oci.oci_image_manifest import OCIImageManifest, ContentDescriptor
from olot.oci.oci_common import MediaTypes

DOCKER_LIST_V2 = "application/vnd.docker.distribution.manifest.list.v2+json"
DOCKER_MANIFEST_V2 = "application/vnd.docker.distribution.manifest.v2+json"
DOCKER_LAYER_TAR_GZIP = "application/vnd.docker.image.rootfs.diff.tar.gzip"
DOCKER_CONFIG_V1 = "application/vnd.docker.container.image.v1+json"

logger = logging.getLogger(__name__)

def check_if_oci_layout_contains_docker_manifests(directory: Path) -> bool:
    """
    Check if the OCI layout contains Docker distribution manifests.
    """
    blobs_path = directory / "blobs" / "sha256"
    for blob in blobs_path.iterdir():
        if not blob.is_file():  # although not expecting the scenario based on spec.
            continue
        try:
            with open(blob, 'r') as f:
                data = json.load(f)
                if data.get("mediaType") == DOCKER_MANIFEST_V2:
                    return True
        except Exception:
            continue
    return False


def convert_docker_manifests_to_oci(directory: Path) -> Dict[str, str]:
    """
    Scan directory for Docker distribution manifests and convert them to OCI format.
    
    Args:
        directory: Path to directory containing Docker distribution manifests
        
    Returns:
        OCIImageManifest: Converted OCI image manifest
        
    Raises:
        FileNotFoundError: If no Docker manifests are found
        ValueError: If manifest format is invalid or layers have wrong media type
    """
    blobs_path = directory / "blobs" / "sha256"
    img_manifest_files = []
    for blob in blobs_path.iterdir():
        try:
            with open(blob, 'r') as f:
                data = json.load(f)
                if data.get("mediaType") == DOCKER_MANIFEST_V2:
                    img_manifest_files.append(blob)
        except Exception:
            continue
    
    if not img_manifest_files:
        raise FileNotFoundError(f"No Docker distribution manifests found in {directory}")

    converted = {f.name: convert_docker_manifest_to_oci(f, directory) for f in img_manifest_files}

    list_manifest_files = []
    for blob in blobs_path.iterdir():
        try:
            with open(blob, 'r') as f:
                data = json.load(f)
                if data.get("mediaType") == DOCKER_LIST_V2:
                    list_manifest_files.append(blob)
        except Exception:
            continue
    for blob_file in list_manifest_files:
        with open(blob_file, 'r') as file_handle:
            index = OCIImageIndex.model_validate_json(file_handle.read())
            for manifest in index.manifests:
                if manifest.mediaType == DOCKER_MANIFEST_V2:
                    new_digest = converted[manifest.digest.removeprefix("sha256:")]
                    manifest.digest = "sha256:" + new_digest
                    manifest.size = os.stat(blobs_path / new_digest).st_size
                    manifest.mediaType = MediaTypes.manifest
                else:
                    raise ValueError(f"Expected manifest mediaType {DOCKER_MANIFEST_V2}, got {manifest.mediaType}")  # TODO: not implemented scenario
            index.mediaType = MediaTypes.index
            index_json = index.model_dump_json(exclude_none=True)
            new_index_hash = compute_hash_of_str(index_json)
            (blobs_path / new_index_hash).write_text(index_json)
            converted[blob_file.name] = new_index_hash

    index = OCIImageIndex.model_validate_json((directory / "index.json").read_text())
    for manifest in index.manifests:
        new_digest = converted[manifest.digest.removeprefix("sha256:")]
        manifest.digest = "sha256:" + new_digest
        manifest.size = os.stat(blobs_path / new_digest).st_size
        with open(blobs_path / new_digest, 'r') as f:
            new_media_type = json.load(f).get("mediaType")
            manifest.mediaType = new_media_type
    new_index_json = index.model_dump_json(exclude_none=True)
    (directory / "index.json").write_text(new_index_json)

    for from_dd_hash, to_oci_hash in converted.items():
        logger.info("Docker distribution manifest %s is now at OCI manifest %s", from_dd_hash, to_oci_hash)
    return converted


def convert_docker_manifest_to_oci(manifest_file: Path, directory: Path) -> str:
    with open(manifest_file, 'r') as f:
        docker_manifest_data = json.load(f)

    if docker_manifest_data.get("mediaType") != DOCKER_MANIFEST_V2:
        raise ValueError(f"Expected {DOCKER_MANIFEST_V2}, got {docker_manifest_data.get('mediaType')}")

    oci_layers = []
    for layer in docker_manifest_data.get("layers", []):
        if layer.get("mediaType") != DOCKER_LAYER_TAR_GZIP:
            raise ValueError(f"Expected layer mediaType {DOCKER_LAYER_TAR_GZIP}, got {layer.get('mediaType')}")

        oci_layer = ContentDescriptor(
            mediaType=MediaTypes.layer_gzip,
            size=layer["size"],
            digest=layer["digest"],
            urls=None,
            data=None,
            artifactType=None
        )
        oci_layers.append(oci_layer)

    config_descriptor = docker_manifest_data.get("config", {})
    if config_descriptor.get("mediaType") != DOCKER_CONFIG_V1:
        raise ValueError(f"Expected config mediaType {DOCKER_CONFIG_V1}, got {config_descriptor.get('mediaType')}")

    config_digest = config_descriptor["digest"]
    config_hash = config_digest.replace("sha256:", "")
    blobs_path = directory / "blobs" / "sha256"
    config_file = blobs_path / config_hash
    
    if not config_file.exists():
        raise FileNotFoundError(f"Config file not found: {config_file}")
    
    with open(config_file, 'r') as f:
        docker_config_data = json.load(f)

    oci_config = OCIManifestConfig.model_validate(docker_config_data)
    
    oci_config_json = oci_config.model_dump_json(exclude_none=True)
    new_config_hash = compute_hash_of_str(oci_config_json)
    new_config_digest = f"sha256:{new_config_hash}"
    oci_config_descriptor = ContentDescriptor(
        mediaType=MediaTypes.config,
        size=len(oci_config_json.encode('utf-8')),
        digest=new_config_digest,
        urls=None,
        data=None,
        artifactType=None
    )
    (blobs_path / new_config_hash).write_text(oci_config_json)

    oci_manifest = OCIImageManifest(
        schemaVersion=2,
        mediaType=MediaTypes.manifest,
        artifactType=None,
        config=oci_config_descriptor,
        layers=oci_layers
    )
    oci_manifest_json = oci_manifest.model_dump_json(exclude_none=True)
    oci_manifest_hash = compute_hash_of_str(oci_manifest_json)
    (blobs_path / oci_manifest_hash).write_text(oci_manifest_json)

    return oci_manifest_hash




import datetime
from pathlib import Path
import os
import json
from typing import List

from olot.oci.oci_config import HistoryItem, OCIManifestConfig, Rootfs, Type
from olot.oci.oci_image_manifest import ContentDescriptor, create_oci_image_manifest, create_manifest_layers
from olot.oci.oci_image_layout import ImageLayoutVersion, OCIImageLayout, create_ocilayout
from olot.oci.oci_common import MediaTypes, Values
from olot.oci.oci_image_index import Manifest, OCIImageIndex, create_oci_image_index
from olot.utils.files import MIMETypes, tarball_from_file, targz_from_file, walk_files
from olot.utils.types import compute_hash_of_str

def create_oci_artifact_from_model(source_dir: Path, dest_dir: Path):
    """
    Create an OCI artifact from a model directory.

    Args:
        source_dir: The directory containing the model files.
        dest_dir: The directory to write the OCI artifact to. If None, a directory named 'oci' will be created in the source directory.
    """
    if not source_dir.exists():
        raise NotADirectoryError(f"Input directory '{source_dir}' does not exist.")

    if dest_dir is None:
        dest_dir = source_dir / "oci"
    os.makedirs(dest_dir, exist_ok=True)

    sha256_path = dest_dir / "blobs" / "sha256"
    os.makedirs(sha256_path, exist_ok=True)

    # assume flat structure for source_dir for now
    # TODO: handle subdirectories appropriately
    model_files = [source_dir / Path(f) for f in os.listdir(source_dir) if os.path.isfile(os.path.join(source_dir, f))]

    # Populate blobs directory
    layers = create_blobs(model_files, dest_dir)

    # Create the OCI image manifest
    manifest_layers = create_manifest_layers(model_files, layers)
    artifactType = MIMETypes.mlmodel
    manifest = create_oci_image_manifest(
        artifactType=artifactType,
        layers=manifest_layers,
    )
    manifest_json = json.dumps(manifest.dict(exclude_none=True), indent=4, sort_keys=True)
    manifest_SHA = compute_hash_of_str(manifest_json)
    with open(sha256_path / manifest_SHA, "w") as f:
        f.write(manifest_json)

    # Create the OCI image index
    index = create_oci_image_index(
        manifests = [
            Manifest(
                mediaType=MediaTypes.manifest,
                size=os.stat(sha256_path / manifest_SHA).st_size,
                digest=f"sha256:{manifest_SHA}",
                urls = None,
            )
        ]
    )
    index_json = json.dumps(index.dict(exclude_none=True), indent=4, sort_keys=True)
    with open(dest_dir / "index.json", "w") as f:
        f.write(index_json)

    # Create the OCI-layout file
    oci_layout = create_ocilayout()
    with open(dest_dir / "oci-layout", "w") as f:
        f.write(json.dumps(oci_layout.model_dump(), indent=4, sort_keys=True))

    # Create empty config file with digest as name
    empty_config: dict[str, str] = {}
    empty_digest_split = Values.empty_digest.split(":")
    if len(empty_digest_split) == 2:
        with open(dest_dir / "blobs" / "sha256" / empty_digest_split[1], "w") as f:
            f.write(json.dumps(empty_config))
    else:
        raise ValueError(f"Invalid empty_digest format: {Values.empty_digest}")


def create_blobs(model_files: List[Path], dest_dir: Path):
    """
    Create the blobs directory for an OCI artifact.
    """
    layers = {} # layer digest : (precomp, postcomp)
    sha256_path = dest_dir / "blobs" / "sha256"

    for model_file in model_files:
        file_name = os.path.basename(os.path.normpath(model_file))
        # handle model card file if encountered - assume README.md is the modelcard
        if file_name.endswith("README.md"):
            new_layer = targz_from_file(model_file, sha256_path)
            postcomp_chksum = new_layer.layer_digest
            precomp_chksum = new_layer.diff_id     
            layers[file_name] = (precomp_chksum, postcomp_chksum)
        else:
            new_layer = tarball_from_file(model_file, sha256_path)
            checksum = new_layer.layer_digest
            layers[file_name] = (checksum, "")
    return layers


def create_simple_oci_artifact(source_path: Path, oci_layout_path: Path):
    """
    Create a simple OCI artifact from a source directory.
    """
    if not source_path.is_dir():
        raise NotADirectoryError(f"Input directory {str(source_path)!r} does not exist.")
    if not oci_layout_path.is_dir():
        raise NotADirectoryError(f"Output directory '{str(oci_layout_path)!r}' does not exist.")
    
    walked_files = walk_files(source_path)

    blobs_path = oci_layout_path / "blobs" / "sha256"
    blobs_path.mkdir(parents=True, exist_ok=True)

    mc = OCIManifestConfig(os="unknown",
                           architecture="unknown",
                           **{"os.version": None, "os.features": None},
                           rootfs=Rootfs(type=Type.layers, diff_ids=[]),
                           history=[],
                           )
    cds = []
    for e in walked_files:
        prefix = str(e.parent) + "/" if e.parent != Path(".") else "/"
        new_layer = targz_from_file(source_path / e, blobs_path, prefix=prefix)
        layer_digest = new_layer.layer_digest
        layer_stat = os.stat(blobs_path / layer_digest)
        size = layer_stat.st_size
        ctime = layer_stat.st_ctime
        title = prefix + e.name if prefix != "/" else e.name
        la = {"olot.title": title}  # cannot use org.opencontainers.image.title to avoid oras not untarring the blob :(
        cd = ContentDescriptor(
            mediaType=MediaTypes.layer_gzip,
            digest="sha256:"+layer_digest,
            size=size,
            urls=None,
            data=None,
            artifactType=None,
            annotations=la
        )
        mc.rootfs.diff_ids.append("sha256:"+new_layer.diff_id)
        hi = HistoryItem(
            created=datetime.datetime.fromtimestamp(ctime, tz=datetime.timezone.utc).isoformat(timespec="seconds").replace("+00:00","Z"),
            created_by="olot create_simple_oci_artifact "+title,
        )
        if mc.history is not None:
            mc.history.append(hi)
        cds.append(cd)
    mc_json = mc.model_dump_json(exclude_none=True)
    mc_json_hash = compute_hash_of_str(mc_json)
    (blobs_path / mc_json_hash).write_text(mc_json)
    manifest = create_oci_image_manifest(
        config=ContentDescriptor(
                mediaType=MediaTypes.config,
                digest="sha256:" + mc_json_hash,
                size=os.stat(blobs_path / mc_json_hash).st_size,
                urls=None,
                data=None,
                artifactType=None,
            ),
        artifactType="application/json",  # TODO: maybe place here something specific to lmeh
        layers=cds
        )
    manifest_json = manifest.model_dump_json(indent=2, exclude_none=True)
    manifest_SHA = compute_hash_of_str(manifest_json)
    manifest_blob_path = blobs_path / manifest_SHA
    manifest_blob_path.write_text(manifest.model_dump_json(indent=2, exclude_none=True))
    
    layout = OCIImageLayout(imageLayoutVersion=ImageLayoutVersion.field_1_0_0)
    (oci_layout_path / "oci-layout").write_text(layout.model_dump_json(indent=2, exclude_none=True))

    index = OCIImageIndex(schemaVersion=2,
                          mediaType=MediaTypes.index,
                          manifests=[
                              Manifest(mediaType=MediaTypes.manifest,
                                       size=os.stat(manifest_blob_path).st_size,
                                       digest="sha256:"+manifest_SHA,
                                       annotations={"org.opencontainers.image.ref.name": "latest"},
                                       urls=None,
                                       )
                          ],
                          artifactType=None,
                          )
    (oci_layout_path / "index.json").write_text(index.model_dump_json(indent=2, exclude_none=True))



# ArtifactTypeModelManifest specifies the artifact type for a model manifest.
ARTIFACTTYPEMODELMANIFEST = "application/vnd.cncf.model.manifest.v1+json"

# MediaTypeModelConfig specifies the media type for a model configuration.
MEDIATYPEMODELCONFIG = "application/vnd.cncf.model.config.v1+json"

# MediaTypeModelWeightRaw is the media type used for an unarchived, uncompressed model weights.
MEDIATYPEMODELWEIGHTRAW = "application/vnd.cncf.model.weight.v1.raw"

# MediaTypeModelWeight is the media type used for model weights.
MEDIATYPEMODELWEIGHT = "application/vnd.cncf.model.weight.v1.tar"

# MediaTypeModelWeightGzip is the media type used for gzipped model weights.
MEDIATYPEMODELWEIGHTGZIP = "application/vnd.cncf.model.weight.v1.tar+gzip"

# MediaTypeModelWeightZstd is the media type used for zstd compressed model weights.
MEDIATYPEMODELWEIGHTZSTD = "application/vnd.cncf.model.weight.v1.tar+zstd"

# MediaTypeModelWeightConfigRaw is the media type used for an unarchived, uncompressed model weights, including files like `tokenizer.json`, `config.json`, etc.
MEDIATYPEMODELWEIGHTCONFIGRAW = "application/vnd.cncf.model.weight.config.v1.raw"

# MediaTypeModelConfig specifies the media type for configuration of the model weights, including files like `tokenizer.json`, `config.json`, etc.
MEDIATYPEMODELWEIGHTCONFIG = "application/vnd.cncf.model.weight.config.v1.tar"

# MediaTypeModelConfigGzip specifies the media type for gzipped configuration of the model weights, including files like `tokenizer.json`, `config.json`, etc.
MEDIATYPEMODELWEIGHTCONFIGGZIP = "application/vnd.cncf.model.weight.config.v1.tar+gzip"

# MediaTypeModelConfigZstd specifies the media type for zstd compressed configuration of the model weights, including files like `tokenizer.json`, `config.json`, etc.
MEDIATYPEMODELWEIGHTCONFIGZSTD = "application/vnd.cncf.model.weight.config.v1.tar+zstd"

# MediaTypeModelDocRaw is the media type used for an unarchived, uncompressed model documentation, including documentation files like `README.md`, `LICENSE`, etc.
MEDIATYPEMODELDOCRAW = "application/vnd.cncf.model.doc.v1.raw"

# MediaTypeModelDoc specifies the media type for model documentation, including documentation files like `README.md`, `LICENSE`, etc.
MEDIATYPEMODELDOC = "application/vnd.cncf.model.doc.v1.tar"

# MediaTypeModelDocGzip specifies the media type for gzipped model documentation, including documentation files like `README.md`, `LICENSE`, etc.
MEDIATYPEMODELDOCGZIP = "application/vnd.cncf.model.doc.v1.tar+gzip"

# MediaTypeModelDocZstd specifies the media type for zstd compressed model documentation, including documentation files like `README.md`, `LICENSE`, etc.
MEDIATYPEMODELDOCZSTD = "application/vnd.cncf.model.doc.v1.tar+zstd"

# MediaTypeModelCodeRaw is the media type used for an unarchived, uncompressed model code, including code artifacts like scripts, code files etc.
MEDIATYPEMODELCODERAW = "application/vnd.cncf.model.code.v1.raw"

# MediaTypeModelCode specifies the media type for model code, including code artifacts like scripts, code files etc.
MEDIATYPEMODELCODE = "application/vnd.cncf.model.code.v1.tar"

# MediaTypeModelCodeGzip specifies the media type for gzipped model code, including code artifacts like scripts, code files etc.
MEDIATYPEMODELCODEGZIP = "application/vnd.cncf.model.code.v1.tar+gzip"

# MediaTypeModelCodeZstd specifies the media type for zstd compressed model code, including code artifacts like scripts, code files etc.
MEDIATYPEMODELCODEZSTD = "application/vnd.cncf.model.code.v1.tar+zstd"

# MediaTypeModelDatasetRaw is the media type used for an unarchived, uncompressed model datasets, including datasets that may be needed throughout the lifecycle of AI/ML models.
MEDIATYPEMODELDATASETRAW = "application/vnd.cncf.model.dataset.v1.raw"

# MediaTypeModelDataset specifies the media type for model datasets, including datasets that may be needed throughout the lifecycle of AI/ML models.
MEDIATYPEMODELDATASET = "application/vnd.cncf.model.dataset.v1.tar"

# MediaTypeModelDatasetGzip specifies the media type for gzipped model datasets, including datasets that may be needed throughout the lifecycle of AI/ML models.
MEDIATYPEMODELDATASETGZIP = "application/vnd.cncf.model.dataset.v1.tar+gzip"

# MediaTypeModelDatasetZstd specifies the media type for zstd compressed model datasets, including datasets that may be needed throughout the lifecycle of AI/ML models.
MEDIATYPEMODELDATASETZSTD = "application/vnd.cncf.model.dataset.v1.tar+zstd"



import os
import shutil
import subprocess
import typing

def is_oras() -> bool :
    return shutil.which("oras") is not None


def oras_pull(base_image: str, dest: typing.Union[str, os.PathLike], params: typing.Sequence[str]=[]):
    if isinstance(dest, os.PathLike):
        dest = str(dest)
    subprocess.run(["oras", "copy", "--to-oci-layout", *params, base_image, dest+":latest"], check=True)
    blobs_dir = os.path.join(dest, "blobs", "sha256")
    for _, _, files in os.walk(blobs_dir):
        for file in files:
            os.chmod(os.path.join(blobs_dir, file), 0o664) # TODO eventually avoid this by refactor manifest change logic



def oras_push(src: typing.Union[str, os.PathLike], oci_ref: str, params: typing.Sequence[str]=[]):
    if isinstance(src, os.PathLike):
        src = str(src)
    return subprocess.run(["oras", "copy", "--from-oci-layout", *params, src+":latest", oci_ref], check=True)



import datetime
from enum import Enum
import logging
import os
from pathlib import Path
from pprint import pformat
import tarfile
from typing import Dict, List, Sequence
import typing

from olot.dockerdist.convert import check_if_oci_layout_contains_docker_manifests, convert_docker_manifests_to_oci
from olot.modelpack.model_config import Model, ModelConfig, ModelDescriptor, ModelFS, Type
from olot.oci.oci_config import HistoryItem, OCIManifestConfig

from olot.oci.oci_image_index import Manifest, OCIImageIndex, Platform, read_ocilayout_root_index
from olot.oci.oci_image_manifest import OCIImageManifest, ContentDescriptor, create_oci_image_manifest
from olot.oci.oci_image_layout import verify_ocilayout
from olot.oci.oci_common import MediaTypes

from olot.utils.files import LayerStats, handle_remove, tarball_from_file, targz_from_file
from olot.utils.types import compute_hash_of_str

from olot.modelpack import const as modelpack_consts

logger = logging.getLogger(__name__)

class CustomStrEnum(str, Enum):
    """To polyfill back to 3.9"""
    @classmethod
    def values(cls) -> Sequence[str]:
        return [e.value for e in cls]
    

class RemoveOriginals(CustomStrEnum):
    """Strategy to be applied when removing original files
    
    default: remove only model weights, configuration, etc.
    all: like default, but also remove the Model CarD"""
    DEFAULT = "default"
    ALL = "all"


def oci_layers_on_top(
        ocilayout: typing.Union[str, os.PathLike],
        model_files: Sequence[os.PathLike],
        modelcard: typing.Union[os.PathLike, None] = None,
        *,
        remove_originals: typing.Union[RemoveOriginals, None] = None,
        add_modelpack: typing.Union[bool, None] = None):
    """
    Add contents to an oci-layout directory as new blob layers

    Args:
        ocilayout: The oci-layout directory of the base image.
        model_files: PathLike array to be added as new blob layers.
        modelcard: PathLike of the README.md of the ModelCarD, will be added as the last layer with compression and annotations. If indicated, it shouldn't be part of model_files.
        remove_originals: whether to remove the original content files after having added the layers, default: None.
        add_modelpack: whether to add a ModelPack manifest to the multi-arch oci-layout only if not already present, default: None.
    """
    if not isinstance(ocilayout, Path):
        ocilayout = Path(ocilayout)
    if remove_originals:
        logger.info("Invoked with remove_originals=%s to delete original contents after adding as a blob layer.", remove_originals.value)

    # Normalize each path in model_files and modelcard to ensure consistency
    model_files = [Path(mf).resolve() for mf in model_files]
    if modelcard is not None:
        modelcard = Path(modelcard).resolve()

    verify_ocilayout(ocilayout)
    if check_if_oci_layout_contains_docker_manifests(ocilayout):
        logger.warning("OCI layout contains Docker distribution manifests, converting them to OCI format")
        convert_docker_manifests_to_oci(ocilayout)
    ocilayout_root_index: OCIImageIndex = read_ocilayout_root_index(ocilayout)
    ocilayout_indexes: Dict[str, OCIImageIndex] = crawl_ocilayout_indexes(ocilayout, ocilayout_root_index)
    ocilayout_manifests: Dict[str, OCIImageManifest] = crawl_ocilayout_manifests(ocilayout, ocilayout_indexes, ocilayout_root_index)
    new_layers: Dict[str, LayerStats] = {} # layer digest : diff_id

    # check configuration is consistent
    add_modelpack = check_and_sanitize_flag_add_modelpack(add_modelpack, ocilayout_indexes, ocilayout_manifests)
    if modelcard in model_files:
        if remove_originals:
            raise ValueError("ModelCard detected in model_files, while remove_originals flag is set; this is not allowed as it would remove the original ModelCard before having a chance of adding it as its proper layer.")
        else:
            logger.warning("ModelCard detected in model_files, this will result in duplicated layers for the ModelCard (negligible, but not optimal).")

    sha256_path = ocilayout / "blobs" / "sha256"
    for model in model_files:
        model = Path(model)
        new_layer = tarball_from_file(model, sha256_path)
        new_layers[new_layer.layer_digest] = new_layer
        if remove_originals:
            handle_remove(model)
    if modelcard is not None:
        new_layer = targz_from_file(Path(modelcard), sha256_path)
        new_layers[new_layer.layer_digest] = new_layer
        if remove_originals == RemoveOriginals.ALL:
            handle_remove(modelcard)

    new_ocilayout_manifests: Dict[str, str] = {}
    for manifest_hash, manifest in ocilayout_manifests.items():
        logger.debug("manifest_hash: %s, manifest.mediaType: %s", manifest_hash, manifest.mediaType)
        config_sha = manifest.config.digest.removeprefix("sha256:")
        mc = None
        with open(ocilayout / "blobs" / "sha256" / config_sha, "r") as cf:
            mc = OCIManifestConfig.model_validate_json(cf.read())
            if mc.history is None:
                mc.history = []
        for new_layer in new_layers.values():
            layer_digest = new_layer.layer_digest
            layer_stat = os.stat(ocilayout / "blobs" / "sha256" / layer_digest)
            size = layer_stat.st_size
            ctime = layer_stat.st_ctime
            mt = MediaTypes.layer if layer_digest == new_layer.diff_id else MediaTypes.layer_gzip
            la = {"org.opencontainers.image.title": new_layer.title} # if ever used for OCI artifact, add `unpack` annotations 
            is_modelcard = layer_digest != new_layer.diff_id
            if is_modelcard:
                la["io.opendatahub.modelcar.layer.type"] = "modelcard"
            cd = ContentDescriptor(
                mediaType=mt,
                digest="sha256:"+layer_digest,
                size=size,
                urls=None,
                data=None,
                artifactType=None,
                annotations=la
            )
            mc.rootfs.diff_ids.append("sha256:"+new_layer.diff_id)
            hi = HistoryItem(
                created=datetime.datetime.fromtimestamp(ctime, tz=datetime.timezone.utc).isoformat(timespec="seconds").replace("+00:00","Z"),
                created_by="olot oci_layers_on_top "+new_layer.title+(" (modelcard)" if is_modelcard else "")
            )
            mc.history.append(hi)
            manifest.layers.append(cd)
        mc_json = mc.model_dump_json(exclude_none=True)
        with open(ocilayout / "blobs" / "sha256" / config_sha, "w") as cf:
            cf.write(mc_json)
        mc_json_hash = compute_hash_of_str(mc_json)
        os.rename(ocilayout / "blobs" / "sha256" / config_sha, ocilayout / "blobs" / "sha256" / mc_json_hash)
        logger.debug("Renamed config from: %s to %s", config_sha, mc_json_hash)
        config_sha = mc_json_hash
        manifest.config.digest = "sha256:" + config_sha
        manifest.config.size = os.stat(ocilayout / "blobs" / "sha256" / config_sha).st_size
        if manifest.annotations is None:
            manifest.annotations = {}
        manifest.annotations["io.opendatahub.author"] = "olot"
        if modelcard is not None:
            manifest.annotations["io.opendatahub.layers.modelcard"] = "sha256:"+next(reversed(new_layers.keys())) # identify ModelCarD layer from Image Manifest
        check_manifest(manifest, mc)
        manifest_json = manifest.model_dump_json(exclude_none=True)
        with open(ocilayout / "blobs" / "sha256" / manifest_hash, "w") as cf:
            cf.write(manifest_json)
        manifest_json_hash = compute_hash_of_str(manifest_json)
        os.rename(ocilayout / "blobs" / "sha256" / manifest_hash, ocilayout / "blobs" / "sha256" / manifest_json_hash)
        logger.debug("Renamed Manifest from: %s to %s", manifest_hash, manifest_json_hash)
        new_ocilayout_manifests[manifest_hash] = manifest_json_hash
        manifest_hash = manifest_json_hash
    logger.debug("new_ocilayout_manifests: %s", pformat(new_ocilayout_manifests))
    if add_modelpack:
        modelpack_manifest_hash = add_modelpack_manifest(ocilayout, new_layers)
    
    new_ocilayout_indexes: Dict[str, str] = {}
    for index_hash, index in ocilayout_indexes.items():
        logger.debug("index_hash: %s, index.mediaType: %s", index_hash, index.mediaType)
        for m in index.manifests:
            digest = m.digest.removeprefix("sha256:")
            if digest in new_ocilayout_manifests.keys():
                lookup_new_hash = new_ocilayout_manifests[m.digest.removeprefix("sha256:")]
                logger.info("old manifest %s is now at %s", m.digest, lookup_new_hash)
                m.digest = "sha256:" + lookup_new_hash
                m.size = os.stat(ocilayout / "blobs" / "sha256" / lookup_new_hash).st_size
            else:
                logger.info("manifest %s was unchanged", digest)
        if add_modelpack:
            if modelpack_manifest_hash in [m.digest.removeprefix("sha256:") for m in index.manifests]:
                raise ValueError(f"ModelPack manifest already present in Index {index_hash} while add_modelpack=True")
            else:
                logger.debug("adding ModelPack manifest to the index %s", index_hash)
                index.manifests.append(Manifest(
                    mediaType=MediaTypes.manifest,
                    digest="sha256:"+modelpack_manifest_hash,
                    size=os.stat(ocilayout / "blobs" / "sha256" / modelpack_manifest_hash).st_size,
                    urls=None,
                    platform=Platform( # type: ignore[call-arg]
                        architecture="unknown",
                        os="unknown",
                    ),
                    annotations={"io.opendatahub.author": "olot", "io.opendatahub.modelcar.manifest.type":"modelpack"},
                ))
        index_json = index.model_dump_json(exclude_none=True)
        with open(ocilayout / "blobs" / "sha256" / index_hash, "w") as idxf:
            idxf.write(index_json)
        index_json_hash = compute_hash_of_str(index_json)
        os.rename(ocilayout / "blobs" / "sha256" / index_hash, ocilayout / "blobs" / "sha256" / index_json_hash)
        logger.debug("Renamed Index from: %s to %s", index_hash, index_json_hash)
        new_ocilayout_indexes[index_hash] = index_json_hash
    logger.debug("new_ocilayout_indexes: %s", pformat(new_ocilayout_indexes))
    
    for entry in ocilayout_root_index.manifests:
        if entry.mediaType == MediaTypes.index:
            lookup_new_hash = new_ocilayout_indexes[entry.digest.removeprefix("sha256:")]
            logger.debug("old index %s is now at %s", entry.digest, lookup_new_hash)
            entry.digest = "sha256:" + lookup_new_hash
            entry.size = os.stat(ocilayout / "blobs" / "sha256" / lookup_new_hash).st_size
        elif entry.mediaType == MediaTypes.manifest:
            digest = entry.digest.removeprefix("sha256:")
            if digest in new_ocilayout_manifests.keys():
                lookup_new_hash = new_ocilayout_manifests[entry.digest.removeprefix("sha256:")]
                logger.info("old manifest %s is now at %s", entry.digest, lookup_new_hash)
                entry.digest = "sha256:" + lookup_new_hash
                entry.size = os.stat(ocilayout / "blobs" / "sha256" / lookup_new_hash).st_size
            else:
                logger.info("manifest %s was unchanged", digest)
        else:
            raise ValueError(f"unknown root index mediaType {entry.mediaType}")
    if add_modelpack:
        if modelpack_manifest_hash in [m.digest.removeprefix("sha256:") for m in ocilayout_root_index.manifests]:
            raise ValueError("ModelPack manifest already present in Root index in oci-layout, while add_modelpack=True")
        else:
            logger.debug("adding ModelPack manifest to the root index")
            ocilayout_root_index.manifests.append(Manifest(
                mediaType=MediaTypes.manifest,
                digest="sha256:"+modelpack_manifest_hash,
                size=os.stat(ocilayout / "blobs" / "sha256" / modelpack_manifest_hash).st_size,
                urls=None,
                platform=Platform( # type: ignore[call-arg]
                    architecture="unknown",
                    os="unknown",
                ),
                annotations={"io.opendatahub.author": "olot", "io.opendatahub.modelcar.manifest.type":"modelpack"},
            ))
    with open(ocilayout / "index.json", "w") as root_idx_f:
        root_idx_f.write(ocilayout_root_index.model_dump_json(exclude_none=True))


def add_modelpack_manifest(ocilayout: Path, new_layers: Dict[str, LayerStats]) -> str:
    """add a ModelPack manifest to the oci-layout
    """
    model_config = Model(
        descriptor=ModelDescriptor(name=None), # eventually config and metadata will be provided programmatically
        modelfs=ModelFS(type=Type.layers, diffIds=list(b.diff_id for b in new_layers.values())),
        config=ModelConfig(),
    )
    model_config_json = model_config.model_dump_json(exclude_none=True)
    model_config_hash = compute_hash_of_str(model_config_json)
    with open(ocilayout / "blobs" / "sha256" / model_config_hash, "w") as mpcf:
        mpcf.write(model_config_json)
    cds = []
    for layer in new_layers.values():
        layer_digest = layer.layer_digest
        layer_stat = os.stat(ocilayout / "blobs" / "sha256" / layer_digest)
        size = layer_stat.st_size
        la = {"org.opencontainers.image.title": layer.title}
        is_modelcard = layer_digest != layer.diff_id
        mt = modelpack_consts.MEDIATYPEMODELDOCGZIP if is_modelcard else modelpack_consts.MEDIATYPEMODELWEIGHT
        cd = ContentDescriptor(
            mediaType=mt,
            digest="sha256:"+layer_digest,
            size=size,
            urls=None,
            data=None,
            artifactType=None,
            annotations=la
        )
        cds.append(cd)
    manifest = create_oci_image_manifest(
        mediaType=MediaTypes.manifest, # default already, but for expliciting here Modelpack spec adherence
        artifactType=modelpack_consts.ARTIFACTTYPEMODELMANIFEST,
        config=ContentDescriptor(
            mediaType=modelpack_consts.MEDIATYPEMODELCONFIG,
            artifactType=None,
            digest="sha256:"+model_config_hash,
            size=os.stat(ocilayout / "blobs" / "sha256" / model_config_hash).st_size,
            urls=None,
            data=None,
        ),
        layers=cds
    )
    manifest_json = manifest.model_dump_json(exclude_none=True)
    manifest_hash = compute_hash_of_str(manifest_json)
    with open(ocilayout / "blobs" / "sha256" / manifest_hash, "w") as mf:
        mf.write(manifest_json)
    return manifest_hash


def check_and_sanitize_flag_add_modelpack(add_modelpack: typing.Union[bool, None], ocilayout_indexes: Dict[str, OCIImageIndex], ocilayout_manifests: Dict[str, OCIImageManifest]) -> bool:
    """Check and sanitize the add_modelpack flag
    - check if the oci-layout contains an Index manifest for multi-arch, othewise fail: can't add a ModelPack manifest to a single-arch oci-layout
      this is because a single-arch is a single OCI Image Manifest and not introducing an Index; hence, can't add a ModelPack manifest to the non-existing Index which is tagged (`:latest`) by the oci-layout root index
    - sanitize the add_modelpack flag to avoid adding a ModelPack manifest if it's already present in the oci-layout
    """
    if add_modelpack:
        if len(ocilayout_indexes) == 0:
            raise ValueError("Can't add a ModelPack manifest to a single-arch oci-layout")
        logger.debug("checking if ModelPack manifest is alreadypresent in the oci-layout")
        for manifest_hash, manifest in ocilayout_manifests.items():
            if manifest.artifactType == modelpack_consts.ARTIFACTTYPEMODELMANIFEST:
                logger.warning("ModelPack manifest already present in the oci-layout as %s; will IGNORE add_modelpack=True", manifest_hash)
                return False
    return add_modelpack or False # leave unchanged unless of sanitization requirements above, if None just return False


def check_manifest(manifest: OCIImageManifest, config: OCIManifestConfig):
    """perform some sanity check on the manifests required for additional scenarios of usage
    """
    ch_count = len(list(x for x in config.history if not x.empty_layer)) if config.history else 0
    layers_count = len(manifest.layers)
    if layers_count != ch_count:
        raise ValueError(f"history lists {ch_count} non-empty layers, but there are {layers_count} layers in the image manifest")


def crawl_ocilayout_manifests(ocilayout: Path, ocilayout_indexes: Dict[str, OCIImageIndex], ocilayout_root_index: typing.Union[OCIImageIndex, None] = None) -> Dict[str, OCIImageManifest]:
    """crawl Manifests from referred OCI Index(es) and Manifests in the root index of the oci-layout
    """
    ocilayout_manifests: Dict[str, OCIImageManifest]  = {}
    for _, mi in ocilayout_indexes.items():
        for m in mi.manifests:
            logger.debug("Parsing manifest %s", m)
            if m.mediaType != MediaTypes.manifest:
                raise ValueError("Did not expect something else than Image Manifest in a Index")
            target_hash = m.digest.removeprefix("sha256:")
            logger.debug("target_hash %s", target_hash)
            manifest_path = ocilayout / "blobs" / "sha256" / target_hash
            with open(manifest_path, "r") as ip:
                ocilayout_manifests[target_hash] = OCIImageManifest.model_validate_json(ip.read())
    for m in ocilayout_root_index.manifests if ocilayout_root_index is not None else []:
        if m.mediaType == MediaTypes.manifest:
            target_hash = m.digest.removeprefix("sha256:")
            logger.debug("Lookup remainder manifest from ocilayout_root_index having target_hash %s", target_hash)
            manifest_path = ocilayout / "blobs" / "sha256" / target_hash
            with open(manifest_path, "r") as ip:
                ocilayout_manifests[target_hash] = OCIImageManifest.model_validate_json(ip.read())

    # filter out non-runnable OCI Images, like Vendor'd Attestations format, and log it out
    filtered: Dict[str, OCIImageManifest]  = {}
    for k, v in ocilayout_manifests.items():
        if v.layers[0].mediaType == "application/vnd.in-toto+json" or v.artifactType == "application/vnd.docker.attestation.manifest.v1+json":
            logger.info("skipping %s as it's an Attestation manifest", k) # not adding this to filtered list of manifests.
        else:
            filtered[k] = v

    return filtered


def write_empty_config_in_ocilayoyt(ocilayout: Path):
    """small utility to avoid limitation of skopeo that can't read inline empty config
    """
    blobs_path = ocilayout / "blobs" / "sha256"
    blobs_path.mkdir(parents=True, exist_ok=True)
    with open(blobs_path / "44136fa355b3678a1146ad16f7e8649e94fb4fc21fe77e8310c060f61caaff8a", 'w') as f:
        f.write("{}")


def crawl_ocilayout_indexes(ocilayout: Path, ocilayout_root_index: OCIImageIndex) -> Dict[str, OCIImageIndex] :
    ocilayout_indexes: Dict[str, OCIImageIndex] = {}
    for m in ocilayout_root_index.manifests:
        if m.mediaType == MediaTypes.index:
            target_hash = m.digest.removeprefix("sha256:")
            index_path = ocilayout / "blobs" / "sha256" / target_hash
            with open(index_path, "r") as ip:
                ocilayout_indexes[target_hash] = OCIImageIndex.model_validate_json(ip.read())
    return ocilayout_indexes


def crawl_ocilayout_blobs_to_extract(ocilayout: Path, 
                                     output_path: Path,
                                     tar_filter_dir: str = "/models") -> List[str]:
    """
    Extract from OCI Image/ModelCar only the contents from a specific directory.

    Args:
        ocilayout: The directory containing the oci-layout of the OCI Image/ModelCar.
        output_path: The directory where to extract the ML model assets from the ModelCar to.
        tar_filter_dir: The subdirectory in the ModelCar to extract, defaults to `"/models"`.

    Returns:
        The list of extracted ML contents from the OCI Image/ModelCar.
    """
    extracted: List[str] = []
    tar_filter_dir= tar_filter_dir.lstrip("/")
    blobs_path = ocilayout / "blobs" / "sha256"
    if not os.path.exists(output_path):
        os.makedirs(output_path)
    verify_ocilayout(ocilayout)
    ocilayout_root_index = read_ocilayout_root_index(ocilayout)
    if len(ocilayout_root_index.manifests) != 1:
        raise ValueError("TODO the root index has more than one manifest, expected single ModelCar")
    manifest0 = ocilayout_root_index.manifests[0]
    if manifest0.mediaType != MediaTypes.manifest:
        raise ValueError("Can only extract from ModelCar Image manifests")
    target_hash = manifest0.digest.removeprefix("sha256:")
    manifest_path = blobs_path / target_hash
    with open(manifest_path, "r") as ip:
        image_manifest = OCIImageManifest.model_validate_json(ip.read())
    for layer in image_manifest.layers:
        if layer.mediaType == MediaTypes.layer or layer.mediaType == MediaTypes.layer_gzip:
            target_hash = layer.digest.removeprefix("sha256:")
            manifest_path = blobs_path / target_hash
            with tarfile.open(manifest_path, "r:*") as tar:
                for member in tar.getmembers():
                    if member.isfile() and member.name.startswith(tar_filter_dir):
                        tar.extract(member, path=output_path)
                        extracted.append(member.name)
    return extracted



"""Modelpack module for handling model artifact configurations."""

from .model_config import Model, ModelConfig, ModelDescriptor, ModelFS, Type, Modality, ModelCapabilities

__all__ = ["Model", "ModelConfig", "ModelDescriptor", "ModelFS", "Type", "Modality", "ModelCapabilities"] 


from os import PathLike
import click
import logging

from .basics import RemoveOriginals, oci_layers_on_top


@click.command()
@click.option("-m", "--modelcard", type=click.Path(exists=True, file_okay=True, dir_okay=False), help="file to be used for ModelCarD; if provided, make sure it's not part of [MODEL_FILES] arguments to avoid redundancies.")
@click.option("--add-modelpack", is_flag=True)
@click.argument('ocilayout', type=click.Path(exists=True, file_okay=False, dir_okay=True))
@click.argument('model_files', nargs=-1)
@click.option('-r', '--remove-originals', type=click.Choice([e.value for e in RemoveOriginals], case_sensitive=False), is_flag=False, flag_value=RemoveOriginals.DEFAULT)
def cli(ocilayout: str, modelcard: PathLike, model_files, remove_originals: bool, add_modelpack: bool):
    logging.basicConfig(level=logging.INFO)
    oci_layers_on_top(ocilayout, model_files, modelcard, remove_originals=RemoveOriginals(remove_originals) if remove_originals else None, add_modelpack=add_modelpack)



# generated by datamodel-codegen:
#   filename:  image-manifest-schema.json
#   timestamp: 2024-12-04T11:34:21+00:00

from __future__ import annotations

from typing import Annotated, List, Optional, Dict
import logging
import os
import subprocess
from pathlib import Path

from pydantic import BaseModel, Field

from olot.oci.oci_common import Urls, Keys, Values, MediaTypes, MediaType
from olot.utils.types import Int64, Base64, Annotations
from olot.utils.files import MIMETypes

logger = logging.getLogger(__name__)

# class MediaType(BaseModel):
#     __root__: constr(
#         regex=r'^[A-Za-z0-9][A-Za-z0-9!#$&^_.+-]{0,126}/[A-Za-z0-9][A-Za-z0-9!#$&^_.+-]{0,126}$'
#     )


# class Digest(BaseModel):
#     __root__: constr(regex=r'^[a-z0-9]+(?:[+._-][a-z0-9]+)*:[a-zA-Z0-9=_-]+$') = Field(
#         ...,
#         description="the cryptographic checksum digest of the object, in the pattern '<algorithm>:<encoded>'",
#     )


# class Urls(BaseModel):
#     __root__: List[AnyUrl] = Field(
#         ..., description='a list of urls from which this object may be downloaded'
#     )


# class MapStringString(BaseModel):
#     __root__: Dict[constr(regex=r'.{1,}'), str]


# class Int8(BaseModel):
#     __root__: conint(ge=-128, le=127)


# class Int16(BaseModel):
#     __root__: conint(ge=-32768, le=32767)


# class Int32(BaseModel):
#     __root__: conint(ge=-2147483648, le=2147483647)


# class Int64(BaseModel):
#     __root__: conint(ge=-9223372036854776000, le=9223372036854776000)


# class Uint8(BaseModel):
#     __root__: conint(ge=0, le=255)


# class Uint16(BaseModel):
#     __root__: conint(ge=0, le=65535)


# class Uint32(BaseModel):
#     __root__: conint(ge=0, le=4294967295)


# class Uint64(BaseModel):
#     __root__: conint(ge=0, le=18446744073709552000)


# class Uint16Pointer(BaseModel):
#     __root__: Optional[Uint16]


# class Uint64Pointer(BaseModel):
#     __root__: Optional[Uint64]


# class Base64(BaseModel):
#     __root__: str


# class StringPointer(BaseModel):
#     __root__: Optional[str]


# class MapStringObject(BaseModel):
#     __root__: Dict[constr(regex=r'.{1,}'), Dict[str, Any]]


# class Annotations(BaseModel):
#     __root__: MapStringString


class ContentDescriptor(BaseModel):
    mediaType: MediaType = Field(
        ..., description="The media type of the referenced object"
    )
    size: Int64 = Field(
        ..., description="The size in bytes of the referenced object"
    )
    digest: str = Field(
        ..., description="The cryptographic checksum digest of the object, in the pattern '<algorithm>:<encoded>'"
    )
    urls: Optional[Urls] = Field(
        None, description="A list of URLs from which this object may be downloaded"
    )
    data: Optional[Base64] = Field(
        None, description="An embedding of the targeted content (base64 encoded)"
    )
    artifactType: Optional[MediaType] = Field(
        None, description="The IANA media type of this artifact"
    )
    annotations: Optional[Dict[str, str]] = None

    class Config:
        exclude_none = True

class OCIImageManifest(BaseModel):
    schemaVersion: Annotated[int, Field(ge=2, le=2)] = Field(
        ...,
        description='This field specifies the image manifest schema version as an integer',
    )
    mediaType: Optional[MediaType] = Field(
        None, description='the mediatype of the referenced object'
    )
    artifactType: Optional[MediaType] = Field(
        None, description='the artifact mediatype of the referenced object'
    )
    config: ContentDescriptor
    subject: Optional[ContentDescriptor] = None
    layers: List[ContentDescriptor] = Field(..., min_length=1)
    annotations: Optional[Annotations] = None


def empty_config() -> ContentDescriptor:
    return ContentDescriptor(
        mediaType=MediaTypes.empty,
        size=2,
        digest=Values.empty_digest,
        data=Values.empty_data,
        urls=None,
        artifactType=None,
    )


def create_oci_image_manifest(
    schemaVersion: int = 2,
    mediaType: Optional[str] = MediaTypes.manifest,
    artifactType: Optional[str] = None,
    config: ContentDescriptor = empty_config(),
    subject: Optional[ContentDescriptor] = None,
    layers: List[ContentDescriptor] = [],
    annotations: Optional[Annotations] = None,
) -> OCIImageManifest:
    return OCIImageManifest(
        schemaVersion=schemaVersion,
        mediaType=mediaType,
        artifactType=artifactType,
        config=config,
        subject=subject,
        layers=layers,
        annotations=annotations,
    )


def get_file_media_type(file_path: os.PathLike) -> str:
    """
    Get the MIME type of a file using the `file` command.
    """
    try:
        result = subprocess.run(['file', '--mime-type', '-b', file_path], stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True)
        mime_type = result.stdout.decode('utf-8').strip()
        return mime_type
    except subprocess.CalledProcessError as e:
        logger.warning("Error occurred while getting MIME type: %s", e)
        return MIMETypes.octet_stream
    except Exception as e:
        logger.warning("Unexpected error: %s", e)
        return MIMETypes.octet_stream


def create_manifest_layers(files: List[Path], blob_layers: dict) -> List[ContentDescriptor]:
    """
    Create a list of ContentDescriptor objects representing the layers of an OCI image manifest.

    Args:
        files (List[os.PathLike]): A list of file paths to be used as layers in the manifest.
    Returns:
        List[ContentDescriptor]: A list of ContentDescriptor objects representing the layers of the manifest
    """
    layers: List[ContentDescriptor] = []
    for file in files:
        precomp, postcomp = blob_layers[os.path.basename(file)]
        file_digest = postcomp if postcomp != "" else precomp
        layer = ContentDescriptor(
            mediaType=get_file_media_type(file),
            size=os.stat(file).st_size,
            digest=f"sha256:{file_digest}",
            annotations= {
                Keys.image_title_annotation: os.path.basename(file)
            },
            urls = None,
            data = None,
            artifactType = None,
        )
        layers.append(layer)
    return layers



from typing import Annotated, List
from pydantic import AnyUrl, Field


MediaType = Annotated[str, Field(
        ...,
        pattern=r'^[A-Za-z0-9][A-Za-z0-9!#$&^_.+-]{0,126}/[A-Za-z0-9][A-Za-z0-9!#$&^_.+-]{0,126}$'
    )]


class Keys:
        image_title_annotation = "org.opencontainers.image.title"
        image_created_annotation = "org.opencontainers.image.created"

class Values:
        empty_digest = "sha256:44136fa355b3678a1146ad16f7e8649e94fb4fc21fe77e8310c060f61caaff8a"
        empty_data = "e30="

class MediaTypes:
        """Constant values from OCI Image Manifest spec

        See also: https://github.com/opencontainers/image-spec/blob/main/media-types.md
        """
        manifest: MediaType = "application/vnd.oci.image.manifest.v1+json"
        index: MediaType = "application/vnd.oci.image.index.v1+json"
        layer: MediaType = "application/vnd.oci.image.layer.v1.tar"
        layer_gzip: MediaType = "application/vnd.oci.image.layer.v1.tar+gzip"
        empty: MediaType = "application/vnd.oci.empty.v1+json"
        config: MediaType = "application/vnd.oci.image.config.v1+json"

Digest = Annotated[str, Field(
        ...,
        pattern=r'^[a-z0-9]+(?:[+._-][a-z0-9]+)*:[a-zA-Z0-9=_-]+$',
        description="the cryptographic checksum digest of the object, in the pattern '<algorithm>:<encoded>'",
    )]


Urls = Annotated[List[AnyUrl],Field(
        ..., description='a list of urls from which this object may be downloaded'
    )]





# generated by datamodel-codegen:
#   filename:  image-index-schema.json
#   timestamp: 2024-12-04T08:16:48+00:00

from __future__ import annotations

from typing import Annotated, List, Optional
from pathlib import Path

from pydantic import BaseModel, Field

from olot.oci.oci_common import MediaTypes, MediaType, Digest, Urls
from olot.utils.types import  Int64, Base64, Annotations

class Platform(BaseModel):
    architecture: str
    os: str
    os_version: Optional[str] = Field(None, alias='os.version')
    os_features: Optional[List[str]] = Field(None, alias='os.features')
    variant: Optional[str] = None


# class MediaType(BaseModel):
#     __root__: constr(
#         regex=r'^[A-Za-z0-9][A-Za-z0-9!#$&^_.+-]{0,126}/[A-Za-z0-9][A-Za-z0-9!#$&^_.+-]{0,126}$'
#     )
# class MediaType(RootModel[str]):
#     root: str = Field(
#         ...,
#         pattern=r'^[A-Za-z0-9][A-Za-z0-9!#$&^_.+-]{0,126}/[A-Za-z0-9][A-Za-z0-9!#$&^_.+-]{0,126}$'
#     )


# class Digest(BaseModel):
#     __root__: constr(regex=r'^[a-z0-9]+(?:[+._-][a-z0-9]+)*:[a-zA-Z0-9=_-]+$') = Field(
#         ...,
#         description="the cryptographic checksum digest of the object, in the pattern '<algorithm>:<encoded>'",
#     )
# class Digest(RootModel[str]):
#     root: str = Field(
#         ...,
#         pattern=r'^[a-z0-9]+(?:[+._-][a-z0-9]+)*:[a-zA-Z0-9=_-]+$',
#         description="the cryptographic checksum digest of the object, in the pattern '<algorithm>:<encoded>'",
#     )


# class Urls(BaseModel):
#     __root__: List[AnyUrl] = Field(
#         ..., description='a list of urls from which this object may be downloaded'
#     )
# class Urls(RootModel[List[AnyUrl]]):
#     root: List[AnyUrl] = Field(
#         ..., description='a list of urls from which this object may be downloaded'
#     )


# class MapStringString(BaseModel):
#     __root__: Dict[constr(regex=r'.{1,}'), str]
# class MapStringString(RootModel[Dict[constr(pattern=r".{1,}"), str]]):
#     """
#     A Pydantic RootModel for a dictionary where keys are non-empty strings
#     and values are strings.
#     """
#     root: Dict[constr(pattern=r".{1,}"), str] = Field(
#         ...,
#         description="A dictionary where keys are non-empty strings and values are strings."
#     )


# class Int8(BaseModel):
#     __root__: conint(ge=-128, le=127)
# class Int8(RootModel[conint(ge=-128, le=127)]):
#     root: conint(ge=-128, le=127) = Field(
#         ...,
#         description="An integer constrained to the 8-bit signed range (-128 to 127)."
#     )


# class Int16(BaseModel):
#     __root__: conint(ge=-32768, le=32767)


# class Int32(BaseModel):
#     __root__: conint(ge=-2147483648, le=2147483647)


# class Int64(BaseModel):
#     __root__: conint(ge=-9223372036854776000, le=9223372036854776000)
# class Int64(RootModel[int]):
#     root: int = Field(
#         ...,
#         ge=-9223372036854776000,
#         le=9223372036854776000,
#     )

# class Uint8(BaseModel):
#     __root__: conint(ge=0, le=255)


# class Uint16(BaseModel):
#     __root__: conint(ge=0, le=65535)


# class Uint32(BaseModel):
#     __root__: conint(ge=0, le=4294967295)


# class Uint64(BaseModel):
#     __root__: conint(ge=0, le=18446744073709552000)
# class Uint64(RootModel[int]):
#     root: int = Field(
#         ...,
#         ge=0,
#         le=18446744073709552000,
#     )


# class Uint16Pointer(BaseModel):
#     __root__: Optional[Uint16]


# class Uint64Pointer(BaseModel):
#     __root__: Optional[Uint64]


# class Base64(RootModel[str]):
#     pass


# class StringPointer(BaseModel):
#     __root__: Optional[str]


# class MapStringObject(BaseModel):
#     __root__: Dict[constr(regex=r'.{1,}'), Dict[str, Any]]


# class Annotations(BaseModel):
#     __root__: MapStringString
# class Annotations(RootModel[MapStringString]):
#     pass


class ContentDescriptor(BaseModel):
    mediaType: MediaType = Field(
        ..., description='the mediatype of the referenced object'
    )
    size: Int64 = Field(..., description='the size in bytes of the referenced object')
    digest: Digest = Field(
        ...,
        description="the cryptographic checksum digest of the object, in the pattern '<algorithm>:<encoded>'",
    )
    urls: Optional[Urls] = Field(
        None, description='a list of urls from which this object may be downloaded'
    )
    data: Optional[Base64] = Field(
        None, description='an embedding of the targeted content (base64 encoded)'
    )
    artifactType: Optional[MediaType] = Field(
        None, description='the IANA media type of this artifact'
    )
    annotations: Optional[Annotations] = None


class Manifest(BaseModel):
    mediaType: MediaType = Field(
        ..., description='the mediatype of the referenced object'
    )
    size: Int64 = Field(..., description='the size in bytes of the referenced object')
    digest: Digest = Field(
        ...,
        description="the cryptographic checksum digest of the object, in the pattern '<algorithm>:<encoded>'",
    )
    urls: Optional[Urls] = Field(
        None, description='a list of urls from which this object may be downloaded'
    )
    platform: Optional[Platform] = None
    annotations: Optional[Annotations] = None


class OCIImageIndex(BaseModel):
    schemaVersion: Annotated[int, Field(ge=2, le=2)] = Field(
        ...,
        description='This field specifies the image index schema version as an integer',
    )
    mediaType: Optional[MediaType] = Field(
        None, description='the mediatype of the referenced object'
    )
    artifactType: Optional[MediaType] = Field(
        None, description='the artifact mediatype of the referenced object'
    )
    subject: Optional[ContentDescriptor] = None
    manifests: List[Manifest]
    annotations: Optional[Annotations] = None


def read_ocilayout_root_index(ocilayout: Path) -> OCIImageIndex:
    ocilayout_root_index = None
    with open(ocilayout / "index.json", "r") as f:
        ocilayout_root_index = OCIImageIndex.model_validate_json(f.read())
    return ocilayout_root_index


def create_oci_image_index(
    schemaVersion: int = 2,
    mediaType: Optional[str] = MediaTypes.index,
    artifactType: Optional[str] = None,
    subject: Optional[ContentDescriptor] = None,
    manifests: List[Manifest] = [],
    annotations: Optional[Annotations] = None
) -> OCIImageIndex:
    """
    Create an OCI image index object.
    """
    return OCIImageIndex(
        schemaVersion=schemaVersion,
        mediaType=mediaType,
        artifactType=artifactType,
        subject=subject,
        manifests=manifests,
        annotations=annotations
    )


# generated by datamodel-codegen using https://github.com/modelpack/model-spec/blob/8f6beb752394e5c6653cfcb3d2cd21950296dfac/schema/config-schema.json
#   filename:  config-schema.json
#   timestamp: 2025-07-04T09:20:40+00:00

from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import List, Optional

from pydantic import BaseModel, ConfigDict, Field


class ModelDescriptor(BaseModel):
    model_config = ConfigDict(extra='forbid')

    createdAt: Optional[datetime] = None
    authors: Optional[List[str]] = None
    family: Optional[str] = None
    name: Optional[str] = Field(None, min_length=1)
    docURL: Optional[str] = None
    sourceURL: Optional[str] = None
    version: Optional[str] = None
    revision: Optional[str] = None
    vendor: Optional[str] = None
    licenses: Optional[List[str]] = None
    title: Optional[str] = None
    description: Optional[str] = None


class Type(Enum):
    layers = 'layers'


class ModelFS(BaseModel):
    model_config = ConfigDict(extra='forbid')

    type: Type
    diffIds: List[str] = Field(..., min_length=1)


class Modality(Enum):
    text = 'text'
    image = 'image'
    audio = 'audio'
    video = 'video'
    embedding = 'embedding'
    other = 'other'


class ModelCapabilities(BaseModel):
    inputTypes: Optional[List[Modality]] = None
    outputTypes: Optional[List[Modality]] = None
    knowledgeCutoff: Optional[datetime] = None
    reasoning: Optional[bool] = None
    toolUsage: Optional[bool] = None


class ModelConfig(BaseModel):
    model_config = ConfigDict(extra='forbid')

    architecture: Optional[str] = None
    format: Optional[str] = None
    paramSize: Optional[str] = None
    precision: Optional[str] = None
    quantization: Optional[str] = None
    capabilities: Optional[ModelCapabilities] = None


class Model(BaseModel):
    model_config = ConfigDict(extra='forbid')

    descriptor: ModelDescriptor
    modelfs: ModelFS
    config: ModelConfig



# generated by datamodel-codegen:
#   filename:  defs.json
#   timestamp: 2024-12-05T10:30:59+00:00

from __future__ import annotations

from typing import List, Optional

from pydantic import BaseModel, Field


class Int8(BaseModel):
    type: str
    minimum: int
    maximum: int


class Int16(BaseModel):
    type: str
    minimum: int
    maximum: int


class Int32(BaseModel):
    type: str
    minimum: int
    maximum: int


class Int64(BaseModel):
    type: str
    minimum: int
    maximum: int


class Uint8(BaseModel):
    type: str
    minimum: int
    maximum: int


class Uint16(BaseModel):
    type: str
    minimum: int
    maximum: int


class Uint32(BaseModel):
    type: str
    minimum: int
    maximum: int


class Uint64(BaseModel):
    type: str
    minimum: int
    maximum: int


class OneOfItem(BaseModel):
    field_ref: Optional[str] = Field(None, alias='$ref')
    type: Optional[str] = None


class Uint16Pointer(BaseModel):
    oneOf: List[OneOfItem]


class Uint64Pointer(BaseModel):
    oneOf: List[OneOfItem]


class Media(BaseModel):
    binaryEncoding: str


class Base64(BaseModel):
    type: str
    media: Media


class OneOfItem2(BaseModel):
    type: str


class StringPointer(BaseModel):
    oneOf: List[OneOfItem2]


class Field1(BaseModel):
    type: str


class PatternProperties(BaseModel):
    field__1__: Field1 = Field(..., alias='.{1,}')


class MapStringString(BaseModel):
    type: str
    patternProperties: PatternProperties


class PatternProperties1(BaseModel):
    field__1__: Field1 = Field(..., alias='.{1,}')


class MapStringObject(BaseModel):
    type: str
    patternProperties: PatternProperties1


class Definitions(BaseModel):
    int8: Int8
    int16: Int16
    int32: Int32
    int64: Int64
    uint8: Uint8
    uint16: Uint16
    uint32: Uint32
    uint64: Uint64
    uint16Pointer: Uint16Pointer
    uint64Pointer: Uint64Pointer
    base64: Base64
    stringPointer: StringPointer
    mapStringString: MapStringString
    mapStringObject: MapStringObject


class Model(BaseModel):
    description: str
    definitions: Definitions



from dataclasses import dataclass
import hashlib
import logging
import shutil
import tarfile
from pathlib import Path
import gzip
import os
from typing import List

logger = logging.getLogger(__name__)

class HashingWriter:
    def __init__(self, base_writer, hash_func=None):
        self.base_writer = base_writer
        self.hash_func = hash_func or hashlib.sha256()

    def write(self, data: bytes):
        self.hash_func.update(data)
        return self.base_writer.write(data)

    def tell(self):
        return self.base_writer.tell()

    def close(self):
        self.base_writer.close()

class MIMETypes:
    mlmodel = "application/x-mlmodel"
    octet_stream = "application/octet-stream"

def get_file_hash(path) -> str:
    h = hashlib.sha256()
    with open(path, "rb") as f:
        while chunk := f.read(4096):
            h.update(chunk)
    return h.hexdigest()


@dataclass
class LayerStats:
    layer_digest: str
    diff_id: str # will be same as layer_digest if only tar, not targz
    title: str


def tar_filter_fn(input: tarfile.TarInfo) -> tarfile.TarInfo :
    """
    A filter function for modifying file metadata when adding files to a tar archive.

    See also: https://docs.openshift.com/container-platform/4.17/openshift_images/create-images.html#use-uid_create-images

    Args:
        input (tarfile.TarInfo): The file metadata object representing a file being added to the tar archive.

    Returns:
        tarfile.TarInfo: The modified file metadata with the following changes:
            - `uid` set to 0 (root user).
            - `gid` set to 0 (root group).
            - `mode` set to 0o664 (read/write for root owner and root group, read-only for others).
    """
    input.uid = 0
    input.gid = 0
    input.mode = 0o664
    return input


def tarball_from_file(file_path: Path, dest: Path, prefix: str = "/models/") -> LayerStats:
    """
    Creates a tarball from a specified file, storing it in the destination directory with a name based on its checksum.

    Args:
        file_path (Path): The path of the file or directory to be added to the tar archive.
        dest (Path): The destination directory where the tarball will be saved.
        prefix (str): The prefix path for the file_path in the tar. Defaults to `/models/` (KServe ModelCar).

    Returns:
        LayerStats: including the checksum (SHA-256) of the tarball, which is also used as the filename.

    Raises:
        ValueError: If supplied input file path does not exist.
        OSError: If an error occurs during file operations such as renaming or writing.
        tarfile.TarError: If an error occurs during the tar archive creation process.
    """
    if not file_path.exists():
        raise ValueError(f"Input file '{file_path}' does not exist.")
    if not prefix.endswith('/'):
        raise ValueError(f"Supplied prefix {prefix} should end with '/'.")
    
    os.makedirs(dest, exist_ok=True)
    temp_dest = dest / "temp"

    try:
        with open(temp_dest, "wb") as temp_file:
            writer = HashingWriter(temp_file)
            with tarfile.open(fileobj=writer, mode="w") as tar: # type: ignore[call-overload]
                tar.add(file_path, arcname=prefix+file_path.name, filter=tar_filter_fn)
        checksum = writer.hash_func.hexdigest()
        os.rename(temp_dest, dest / checksum)
        return LayerStats(checksum, checksum, file_path.name)
    except tarfile.TarError as e:
        raise tarfile.TarError(f"Error creating tarball: {e}") from e
    except OSError as e:
        raise OSError(f"File operation failed: {e}") from e


def targz_from_file(file_path: Path, dest: Path, prefix: str = "/models/") -> LayerStats:
    """
    Creates a gzipped tarball from the specified file, storing it in the destination directory.
    The tarball's filename is based on the post-compression checksum.

    Args:
        file_path (Path): The path of the file or directory to be compressed.
        dest (Path): The destination directory where the gzipped tarball will be saved. If the directory
                     does not exist, it will be created.
        prefix (str): The prefix path for the file_path in the tar. Defaults to `/models/` (KServe ModelCar).

    Returns:
        LayerStats: containing:
            - The post-compression checksum (SHA-256 of the gzipped tarball).
            - The pre-compression checksum (SHA-256 of the tar content).

    Raises:
        ValueError: If the input file or directory does not exist.
        OSError: If a file operation (e.g., writing, renaming) fails.
        tarfile.TarError: If an error occurs during tarball creation.
    """
    if not file_path.exists():
        raise ValueError(f"Input file '{file_path}' does not exist.")
    if not prefix.endswith('/'):
        raise ValueError(f"Supplied prefix {prefix} should end with '/'.")

    os.makedirs(dest, exist_ok=True)
    temp_dest = dest / "temp"

    try:
        with open(temp_dest, "wb") as temp_file:
            writer = HashingWriter(temp_file)
            with gzip.GzipFile(fileobj=writer, mode="wb", mtime=0, compresslevel=6) as gz: # type: ignore[call-overload]
                inner_writer = HashingWriter(gz)
                with tarfile.open(fileobj=inner_writer, mode="w") as tar: # type: ignore[call-overload]
                    tar.add(file_path, arcname=prefix+file_path.name, filter=tar_filter_fn)
        precompress_checksum = inner_writer.hash_func.hexdigest()
        postcompress_checksum = writer.hash_func.hexdigest()
        os.rename(temp_dest, dest / postcompress_checksum)
        return LayerStats(postcompress_checksum, precompress_checksum, file_path.name)
    except tarfile.TarError as e:
        raise tarfile.TarError(f"Error creating tarball: {e}") from e
    except OSError as e:
        raise OSError(f"File operation failed: {e}") from e


def handle_remove(path: os.PathLike):
    if not isinstance(path, Path):
        path = Path(path)
    if path.is_symlink():
        logger.warning("removing %s which is a symlink", path)
    if path.is_dir():
        shutil.rmtree(path)
    else:
        os.remove(path)


def walk_files(root_path: os.PathLike) -> List[Path]:
    """
    Recursively walks a directory and returns all files as relative paths, skipping any symlinks.
    
    Args:
        root_path: The root directory to walk recursively
        
    Returns:
        List of relative file paths as strings
        
    Raises:
        ValueError: If the provided path doesn't exist or isn't a directory
        OSError: If an error occurs during directory traversal
    """
    if not isinstance(root_path, Path):
        root_path = Path(root_path)
    
    if not root_path.exists():
        raise FileNotFoundError(f"Path '{root_path}' does not exist")
    
    if not root_path.is_dir():
        raise NotADirectoryError(f"Path '{root_path}' is not a directory")
    
    try:
        relative_files = []
        for dirpath, dirnames, filenames in os.walk(str(root_path)):
            # Skip symlink directories by removing them from dirnames
            dirnames[:] = [d for d in dirnames if not os.path.islink(os.path.join(dirpath, d))]
            
            for filename in filenames:
                full_file_path = os.path.join(dirpath, filename)
                # Skip symlink files
                if not os.path.islink(full_file_path):
                    file_path = Path(full_file_path)
                    relative_path = file_path.relative_to(root_path)
                    relative_files.append(relative_path)
        
        return sorted(relative_files)  # Return sorted for consistent ordering
    except OSError as e:
        raise OSError(f"Error walking directory '{root_path}': {e}") from e
